"""
Sub      : instancing based on a list of coordinates and angles
Ref      : https://www.klayout.de/forum/discussion/1647/instancing-based-on-a-list-of-coordinates-and-angles
Posted by: klaybur
       on: 2020-09-25
Modified by: klaybur
       on: 2020-09-28
File         : 2809_Coord_Angle.py
Author       : Kazzz-S's initial script modified by klaybur
Last modified: 2020-09-28

Notes:
  1) I assume you run this script in the "Macro Development" window.
  2) Tested with KLayout version 0.26.8 on macOS Catalina.
  3) For "Pandas" module, Windows users should refer to https://www.klayout.de/forum/discussion/1645/pandas-library-for-windows
  4) Also refer to:
      a) https://www.klayout.de/forum/discussion/1642/python-scripting-for-reading-an-excel-file-and-creating-a-polar-array-on-klayout-macro
      b) https://github.com/KLayout/klayout/issues/602

Usage:
  (1) Store this script in ${HOME}/.klayout/pymacros/
  (2) Start the Macro Development IDE tool
  (3) Open this script and run
  (4) Load (drag & drop) "KL1647.lyp" layer property file      .
"""
import os
import pandas as pd
import pya

"""
<=== [Kazzz-S] indicates my comments.
     !!! Due to these comments, this script DOES NOT RUN !!!
"""

#-------------------------------------------------------------------------------------
# Global control parameters; modify and set these parameters based on your environment
#-------------------------------------------------------------------------------------
Debug   = True   # set False not to print debug messages
MyRoot  = os.environ["HOME"] + "/KLayout/Forum1647/"
MyCSV   = MyRoot + "KazzzS-KL1647.csv" <=== [Kazzz-S] 1st column is expected to be a cell name.
                                            You need to remake this. See, "2809_Coord_Angle.csv"
OutGDS2 = MyRoot + "KL1647.gds" <=== [Kazzz-S] better to change this


#-------------------------------------------------------------------------------
# [1] Using "Pandas" module, read the CSV file and populate a layout dictionary
#     layout dictionary: key=character (=cell_name)
#                        value=(x-pos[um], y-pos[um], angle[degree])-tuple
#-------------------------------------------------------------------------------
layDic = dict() # the layout dictionary
df     = pd.read_csv( MyCSV, comment='#' )
for idx in range(0, len(df) ):
  char  = str  ( df.iloc[idx, 0].strip("'") )
  xpos  = float( df.iloc[idx, 1] )
  ypos  = float( df.iloc[idx, 2] )
  angle = float( df.iloc[idx, 3] )
  layDic[char] = (xpos, ypos, angle)  # ([um], [um], [degree])
print( "" )
print( "### Successfully read:", MyCSV )
if Debug:
  print(df)

#-------------------------------------------------------------------------------
# [2] Create a new layout with an empty "TOP" cell
#
#     Ref. https://www.klayout.de/doc-qt5/programming/database_api.html#h2-18
#
#            [TOP] <~~~ to be created
#
#-------------------------------------------------------------------------------
mainWin    = pya.MainWindow.instance()
layout     = CurrentView.active_cellview().layout() <=== [Kazzz-S] CurrentView is undefined.
test       = layout.cell_by_name("TEST")
cellTop    = layout.create_cell( "TOP" )
myDBU      = 0.001  # data base unit in [um]
layout.dbu = myDBU


  # (B) Retrieve the layout dictionary with respect to the cellname
xpos, ypos, angle = layDic[test] <=== [Kazzz-S] key to the dictionary is expected to be the string "TEST"
                                 <=== [Kazzz-S] layDic does not have such a key entry.
                                 <=== [Kazzz-S] Hence, you need to change the contents of CSV



  # (D) Prepare a complex transformation: read the KLayout Assistance Document carefully.
  #
  #       Signature: [static] new CplxTrans ptr new (double mag, double rot, bool mirrx, double x, double y)
  #       Description: Creates a transformation using magnification, angle, mirror flag and displacement
  #       mag:
  #       The magnification
  #       rot:
  #       The rotation angle in units of degree
  #       mirrx:
  #       True, if mirrored at x axis
  #       x:
  #       The x displacement
  #       y:
  #       The y displacement
  #
  #       The sequence of operations is: magnification, mirroring at x axis, rotation, application of displacement.
  #       Python specific notes:
  #       This method is the default initializer of the object
  #
cplxTrans = pya.CplxTrans( 1.0, angle, False, xpos/myDBU, ypos/myDBU )

  # (E) Create a cell instance
cellIns            = pya.CellInstArray()    # default cell instance
cellIns.cell_index = cellChild.cell_index() # 'set' the cell to refer (what to refer)
                     <=== [Kazzz-S] cellChild is undefined
cellIns.cplx_trans = cplxTrans              # 'set' the complex transformation (how to refer)

  # (F) Insert the cell instance to [TOP]
cellTop.insert(cellIns)

#-------------------------------------------------------------------------------
# [5] Finishing...
#-------------------------------------------------------------------------------
# (A) Take care of missing layers if any
layout_view.add_missing_layers()

# (B) Show and zoom to fit the entire hierarchy
layout_view.max_hier_levels = 2
layout_view.zoom_fit()

# (C) Export the current design to the GDS2 file
sop = pya.SaveLayoutOptions()   # default save option
layout.write( OutGDS2, sop )    # *.gds makes KLayout infer the output format
print( "" )
print( "### Generated:", OutGDS2 )

#--------------
# End of File
#--------------
