"""
Sub      : instancing based on a list of coordinates and angles
Ref      : https://www.klayout.de/forum/discussion/1647/instancing-based-on-a-list-of-coordinates-and-angles
Posted by: klaybur
       on: 2020-09-25
Modified by: klaybur
       on: 2020-09-28

File         : KazzzS-2809_Coord_Angle.py
Author       : Kazzz-S
Description  : Kazzz-S interpreted klaybur's script "2809_Coord_Angle.py" and corrected errors.
             : Refer to "Comments-2809_Coord_Angle.py" for those errors.
Last modified: 2020-09-29

Notes:
  1) I assume you run this script in the "Macro Development" window.
  2) Tested with KLayout version 0.26.8 on macOS Catalina/Linux-Mint/Windows-10.
  3) For "Pandas" module, Windows users should refer to https://www.klayout.de/forum/discussion/1645/pandas-library-for-windows
  4) Also refer to:
      a) https://www.klayout.de/forum/discussion/1642/python-scripting-for-reading-an-excel-file-and-creating-a-polar-array-on-klayout-macro
      b) https://github.com/KLayout/klayout/issues/602

Usage:
  (1) Store this script in ${HOME}/.klayout/pymacros/
  (2) Start the Macro Development IDE tool
  (3) Open this script and run
  (4) Load (drag & drop) "KL1647.lyp" layer property file if you like.
"""
import os
import pandas as pd
import pya

#-------------------------------------------------------------------------------------
# Global control parameters; modify and set these parameters based on your environment
#-------------------------------------------------------------------------------------
Debug    = True   # set False not to print debug messages
MyRoot   = os.environ["HOME"] + "/KLayout/Forum1647/"
MyCSV    = MyRoot + "2809_Coord_Angle.csv"
MyInGDS2 = MyRoot + "2809_Coord_Angle.GDS"
OutGDS2  = MyRoot + "Out_2809_Coord_Angle.GDS"

#-------------------------------------------------------------------------------
# [1] Using "Pandas" module, read the CSV file and populate a layout dictionary
#     layout dictionary: key=character (=cell_name)
#                        value="list" of (x-pos[um], y-pos[um], angle[degree])-tuple
#
#     To refer to the same cell multiple times, the tuple is stored in a list [].
#
#     When you run this script in the "Macro Development" window, you will get
#     [Exception Caught] dialog pops up on line-62 (marked <====).
#     This exception is harmless. Press [OK] to continue.
#-------------------------------------------------------------------------------
layDic = dict() # the layout dictionary
df     = pd.read_csv( MyCSV, comment='#' )
for idx in range(0, len(df) ):
  cell  = str  ( df.iloc[idx, 0].strip("'") )
  xpos  = float( df.iloc[idx, 1] )
  ypos  = float( df.iloc[idx, 2] )
  angle = float( df.iloc[idx, 3] )
  try:
    tupleList = layDic[cell]
  except KeyError: # <==== [Exception Caught] dialog pops up at the first entry of "cell"
    layDic[cell] = [ (xpos, ypos, angle) ]  # ([um], [um], [degree])
    continue
  else: # additional entry
    layDic[cell] = tupleList + [ (xpos, ypos, angle) ] # extend the existing list
print( "" )
print( "### Successfully read:", MyCSV )
if Debug:
  print( "### DataFrame ###" )
  print(df)
  print( "" )
  print( "### Cell Reference Dictionary ###" )
  for cell in layDic.keys():
    tupleList = layDic[cell]
    print( "  Cell[%s] has the cell reference list of %s" % (cell, tupleList) )

#-------------------------------------------------------------------------------
# [2] Create a new layout and import the GDS2 containing [TEST] cell.
#     Then create a new [TOP] cell.
#
#     Ref. https://www.klayout.de/doc-qt5/programming/database_api.html#h2-18
#
#          [TEST] <~~~ to be imported
#           [TOP] <~~~ to be created
#-------------------------------------------------------------------------------
# (A) Prepare a new layout
mainWin = pya.MainWindow.instance()
layout  = mainWin.create_layout(1).layout()

# (B) Read the specified GDS2 that contains cell(s) to refer
layout.read( MyInGDS2 )
myDBU = layout.dbu # database unit; most likely 0.001 [um]

# (C) Take care of missing layers if any
layout_view = mainWin.current_view()
layout_view.add_missing_layers()

# (D) Find a cell concerned, which will be the child cell
#
#     In a practical code, you need to cross-check
#       1) the presence of all child cells that appear in the "layDic" object.
#       2) the absence of the "top" cell to be newly created.
cellIDChild = layout.cell_by_name( "TEST" )

# (E) Create a new cell, which will be the parent cell
cellTop = layout.create_cell( "TOP" )
layout_view.select_cell( cellTop.cell_index(), 0 )

#-------------------------------------------------------------------------------
# [2] Construct the cell hierarchy shown below
#
#          [TOP]
#           |:|
#           |:| <~~~ different transformations are given by "layDic" dictionary
#           |:|
#          [TEST] <~~ this cell is referred multiple times
#
#-------------------------------------------------------------------------------
# (A) Iterate over all child cell candidates even though we have only one in this case...
for cell in layDic.keys():
  # (B) Iterate over all cell references; eight (8) times for [TEST] cell.
  #     Retrieve the layout dictionary with respect to the child cell name.
  for refinfo in layDic[cell]:
    # (C) Unpack cell the reference information ("how to refer")
    xpos, ypos, angle = refinfo

    # (D) Prepare a complex transformation: read the KLayout Assistance Document carefully.
    #
    #       Signature: [static] new CplxTrans ptr new (double mag, double rot, bool mirrx, double x, double y)
    #       Description: Creates a transformation using magnification, angle, mirror flag and displacement
    #       mag:
    #       The magnification
    #       rot:
    #       The rotation angle in units of degree
    #       mirrx:
    #       True, if mirrored at x axis
    #       x:
    #       The x displacement
    #       y:
    #       The y displacement
    #
    #       The sequence of operations is: magnification, mirroring at x axis, rotation, application of displacement.
    #       Python specific notes:
    #       This method is the default initializer of the object
    #
    cplxTrans = pya.CplxTrans( 1.0, angle, False, xpos/myDBU, ypos/myDBU )

    # (E) Create a cell instance
    cellIns            = pya.CellInstArray()  # default cell instance
    cellIns.cell_index = cellIDChild          # 'set' the cell to refer (what to refer)
    cellIns.cplx_trans = cplxTrans            # 'set' the complex transformation (how to refer)

    # (F) Insert the cell instance to [TOP]
    cellTop.insert(cellIns)

#-------------------------------------------------------------------------------
# [3] Finishing...
#-------------------------------------------------------------------------------
# (A) Take care of missing layers if any
layout_view.add_missing_layers()

# (B) Show and zoom to fit the entire hierarchy
layout_view.max_hier_levels = 2
layout_view.zoom_fit()

# (C) Export the current design to the GDS2 file
sop = pya.SaveLayoutOptions()   # default save option
layout.write( OutGDS2, sop )    # *.GDS makes KLayout infer the output format
print( "" )
print( "### Generated:", OutGDS2 )

#--------------
# End of File
#--------------
